from tqdm import tqdm
from .config import get_temporary_directory, load_config
from .utils import get_unique_filename, postprocess_cleanup, unpack_caption
import os, re, requests, shutil, sys, random, ffmpy

userConfig = load_config()
downloadDIR = userConfig['downloadDIR']
tempDIR = get_temporary_directory()

def download_progressive(stream, itag, title, resolution, file_extention, captions, caption_code=None, tempDIR=tempDIR, downloadDIR=downloadDIR):
    global total_filesize, progress_bar
    selected_vdo = stream.get_by_itag(itag)
    total_filesize = selected_vdo.filesize
    progress_bar = tqdm(total=total_filesize, unit='B', unit_scale=True, desc="Downloading Video+Audio")
    random_filename = str(random.randint(1000000000, 9999999999))
    filename = random_filename + '_vdo.' + file_extention
    output_temp_file = os.path.join(tempDIR, filename)
    output_file = os.path.join(downloadDIR, get_unique_filename(title + '_' + resolution + '.' + file_extention)) if not caption_code else os.path.join(downloadDIR, get_unique_filename(title + '_' + resolution + '_' + caption_code + '.' + file_extention))
    selected_vdo.download(output_path=tempDIR, filename=filename)

    if caption_code:
        print(f'Downloading Caption ({caption_code})...')
        caption = captions[caption_code]
        _, caption_lang = unpack_caption(caption)
        caption_file = os.path.join(tempDIR, random_filename + '_cap.srt')
        caption.save_captions(caption_file)
        print('Processing...')
        devnull = open(os.devnull, 'w')
        output_temp_file_with_subs = os.path.join(tempDIR, random_filename + '_merged.' + file_extention)
        ff = ffmpy.FFmpeg(
            inputs={output_temp_file: None},
            outputs={output_temp_file_with_subs: ['-i', caption_file, '-c', 'copy', '-c:s', 'mov_text', '-metadata:s:s:0', f'language={caption_code}', '-metadata:s:s:0', f'title={caption_lang}', '-metadata:s:s:0', f'handler_name={caption_lang}']}
        )
        ff.run(stdout=devnull, stderr=devnull)
        devnull.close()

        shutil.move(output_temp_file_with_subs, output_file)
        postprocess_cleanup(tempDIR, ['_vdo.' + file_extention, '_cap.srt', '_merged.' + file_extention], random_filename)
        print('Done! 🎉')
    else:
        print('Processing...')
        shutil.move(output_temp_file, output_file)
        print('Done! 🎉')

def download_nonprogressive(stream, itag_vdo, itag_ado, file_extention, output_path):
    global total_filesize, progress_bar
    selected_vdo = stream.get_by_itag(itag_vdo)
    selected_ado = stream.get_by_itag(itag_ado)
    random_filename = str(random.randint(1000000000, 9999999999))
    total_filesize = selected_vdo.filesize
    progress_bar = tqdm(total=total_filesize, unit='B', unit_scale=True, desc="Downloading Video")
    selected_vdo.download(output_path=output_path, filename=random_filename + '_vdo.' + file_extention)
    total_filesize = selected_ado.filesize
    progress_bar = tqdm(total=total_filesize, unit='B', unit_scale=True, desc="Downloading Audio")
    selected_ado.download(output_path=output_path, filename=random_filename + '_ado.' + file_extention)
    return random_filename

def download_audio(stream, itag, output_path):
    global total_filesize, progress_bar
    selected_ado = stream.get_by_itag(itag)
    total_filesize = selected_ado.filesize
    progress_bar = tqdm(total=total_filesize, unit='B', unit_scale=True, desc="Downloading Audio")
    random_filename = str(random.randint(1000000000, 9999999999))
    selected_ado.download(output_path=output_path, filename=random_filename + '_ado.mp4')
    return random_filename

def download_thumbnail(url, file_path):
    print('Downloading thumbnail...')
    maxres_url = re.sub(r'/[^/]*\.jpg.*$', '/maxresdefault.jpg', url)
    hq_url = re.sub(r'/[^/]*\.jpg.*$', '/hqdefault.jpg', url)
    
    response = requests.get(maxres_url, stream=True)
    if response.status_code != 200:
        response = requests.get(hq_url, stream=True)
    
    if response.status_code == 200:
        with open(file_path, 'wb') as file:
            response.raw.decode_content = True
            shutil.copyfileobj(response.raw, file)
    else:
        print('Failed to download thumbnail...!')
        sys.exit()

def progress(chunk, file_handle, bytes_remaining):
    chunk_size = total_filesize - bytes_remaining
    progress_bar.update(chunk_size - progress_bar.n)

    if bytes_remaining == 0:
        progress_bar.close()


import os, json, platform, appdirs, tempfile

def get_download_folder():
    system = platform.system()
    if system in ["Windows", "Darwin", "Linux"]:
        cli_download_dir = os.path.join(os.path.expanduser("~"), "Downloads", "PytubePP Downloads")
        os.makedirs(cli_download_dir, exist_ok=True)
        return cli_download_dir
    else:
        cli_download_dir = os.path.join(appdirs.user_download_dir(), "PytubePP Downloads")
        os.makedirs(cli_download_dir, exist_ok=True)
        return cli_download_dir
    
DEFAULT_CONFIG = {
    'downloadDIR': get_download_folder(),
    'defaultStream': 'max',
    'defaultCaption': 'none',
}
    
def get_temporary_directory():
    temp_dir = tempfile.gettempdir()
    cli_temp_dir = os.path.join(temp_dir, 'pytubepp')
    os.makedirs(cli_temp_dir, exist_ok=True)
    return cli_temp_dir

def load_config():
    config_dir = appdirs.user_config_dir('pytubepp')
    config_path = os.path.join(config_dir, 'config.json')
    if os.path.exists(config_path):
        with open(config_path, 'r') as f:
            return json.load(f)
    else:
        return DEFAULT_CONFIG

def save_config(config):
    config_dir = appdirs.user_config_dir('pytubepp')
    os.makedirs(config_dir, exist_ok=True)
    config_path = os.path.join(config_dir, 'config.json')
    with open(config_path, 'w') as f:
        json.dump(config, f, indent=4)

def update_config(key, value):
    config = load_config()
    config[key] = value
    save_config(config)

def reset_config():
    config_dir = appdirs.user_config_dir('pytubepp')
    config_path = os.path.join(config_dir, 'config.json')
    if os.path.exists(config_path):
        os.remove(config_path)
        print('\nConfig reset successful!')
    else:
        print('\nAlready using the default configs! Not resetting...!')


from .utils import ffmpeg_installed, nodejs_installed
import subprocess, platform

def postinstall():
    os_type = platform.system().lower()
    package_manager = None

    print("### PytubePP Post-Install Script ###\n")

    print("Checking requirements...")
    ffmpeg_needed = not ffmpeg_installed()
    nodejs_needed = not nodejs_installed()

    if ffmpeg_needed or nodejs_needed:
        if os_type == 'windows':
            version_info = platform.version().split('.')
            if int(version_info[0]) >= 10 and (int(version_info[1]) > 0 or int(version_info[2]) >= 1709):
                winget_check = subprocess.run(['winget', '--version'], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
                if winget_check.returncode == 0:
                    print("OS: Windows (winget)")
                    package_manager = 'winget'  # Windows Package Manager
                else:
                    print("OS: Windows (winget not enabled)")
                    user_input = input("WinGet is not available. Do you want to enable winget? (Make sure to login to Windows before enabling) [Yes/no]: ").strip().lower()
                    if user_input in ['yes', 'y', '']:
                        print("Enabling winget...")
                        subprocess.run(['powershell', '-Command', 'Add-AppxPackage -RegisterByFamilyName -MainPackage Microsoft.DesktopAppInstaller_8wekyb3d8bbwe'])
                        print("WinGet enabled successfully! Please restart your computer and re-run the post install script: pytubepp --postinstall")
                        return
                    else:
                        print("Installation aborted! exiting...!!")
                        return
            else:
                print("OS: Windows (winget not supported)")
                print("Unsupported Windows version! WinGet requires Windows 10 1709 (build 16299) or later, Please install dependencies manually...!!")
                return
        elif os_type == 'linux':
            # Determine the Linux distribution
            if subprocess.run(['command', '-v', 'apt'], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL).returncode == 0:
                print("OS: Linux (apt)")
                package_manager = 'apt'  # APT for Debian/Ubuntu
            elif subprocess.run(['command', '-v', 'dnf'], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL).returncode == 0:
                print("OS: Linux (dnf)")
                distro_id = subprocess.run(['grep', '^ID=', '/etc/os-release'], stdout=subprocess.PIPE, stderr=subprocess.DEVNULL)
                if distro_id.returncode == 0 and 'fedora' in distro_id.stdout.decode().strip() and ffmpeg_needed:
                    user_input = input("Looks like you are using Fedora. Do you want to enable RPM Fusion free and nonfree repositories? (answer no if already enabled) [Yes/no]: ").strip().lower()
                    if user_input in ['yes', 'y', '']:
                        print("Enabling RPM Fusion repositories...")
                        fedora_version = subprocess.run(['rpm', '-E', '%fedora'], stdout=subprocess.PIPE, stderr=subprocess.DEVNULL)
                        if fedora_version.returncode == 0:
                            fedora_version_str = fedora_version.stdout.decode().strip()
                            subprocess.run(['sudo', 'dnf', 'install', f'https://download1.rpmfusion.org/free/fedora/rpmfusion-free-release-{fedora_version_str}.noarch.rpm', '-y'], check=True)
                            subprocess.run(['sudo', 'dnf', 'install', f'https://download1.rpmfusion.org/nonfree/fedora/rpmfusion-nonfree-release-{fedora_version_str}.noarch.rpm', '-y'], check=True)
                        else:
                            print("Failed to retrieve Fedora version. Please install RPM Fusion repositories manually.")
                    else:
                        print("RPM Fusion repositories installation skipped...!!")
                package_manager = 'dnf'  # DNF for Fedora
            elif subprocess.run(['command', '-v', 'pacman'], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL).returncode == 0:
                print("OS: Linux (pacman)")
                package_manager = 'pacman'  # Pacman for Arch Linux
            else:
                print("OS: Linux (unknown)")
                print("Unsupported Linux distribution! Please install dependencies manually...!!")
                return
        elif os_type == 'darwin':
            homebrew_check = subprocess.run(['brew', '--version'], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
            if homebrew_check.returncode == 0:
                print("OS: MacOS (brew)")
                package_manager = 'brew'  # Homebrew for macOS
            else:
                print("OS: MacOS (brew not installed)")
                user_input = input("Homebrew is not installed. Do you want to install Homebrew? [Yes/no]: ").strip().lower()
                if user_input in ['yes', 'y', '']:
                    print("Installing Homebrew...")
                    subprocess.run('/bin/bash -c "$(curl -fsSL https://raw.githubusercontent.com/Homebrew/install/HEAD/install.sh)"', shell=True)
                    print("Homebrew installation completed! Please restart your mac and re-run the post install script: pytubepp --postinstall")
                    return
                else:
                    print("Installation aborted! exiting...!!")
                    return
        else:
            print("Unsupported OS! Please install dependencies manually...!!")
            return

        print("The following packages are about to be installed:")
        if ffmpeg_needed:
            print("- FFmpeg")
        if nodejs_needed:
            print("- Node.js")
        
        user_input = input("Do you want to proceed with the installation? [Yes/no]: ").strip().lower()
        if user_input not in ['yes', 'y', '']:
            print("Installation aborted! exiting...!!")
            return

        if ffmpeg_needed:
            print("Installing FFmpeg...")
            install_ffmpeg(package_manager)
        if nodejs_needed:
            print("Installing Node.js...")
            install_nodejs(package_manager)
    else:
        print("Dependencies already satisfied! exiting...!!")

def install_ffmpeg(package_manager):
    if package_manager == 'winget':
        subprocess.run(['winget', 'install', 'ffmpeg'], check=True)
    elif package_manager == 'apt':
        subprocess.run(['sudo', 'apt', 'install', 'ffmpeg', '-y'], check=True)
    elif package_manager == 'dnf':
        subprocess.run(['sudo', 'dnf', 'install', 'ffmpeg', '-y'], check=True)
    elif package_manager == 'pacman':
        subprocess.run(['sudo', 'pacman', '-S', 'ffmpeg', '--noconfirm'], check=True)
    elif package_manager == 'brew':
        subprocess.run(['brew', 'install', 'ffmpeg'], check=True)
    print("FFmpeg installation completed")

def install_nodejs(package_manager):
    if package_manager == 'winget':
        subprocess.run(['winget', 'install', 'OpenJS.NodeJS.LTS'], check=True)
    elif package_manager == 'apt':
        subprocess.run(['sudo', 'apt', 'install', 'nodejs', '-y'], check=True)
    elif package_manager == 'dnf':
        subprocess.run(['sudo', 'dnf', 'install', 'nodejs', '-y'], check=True)
    elif package_manager == 'pacman':
        subprocess.run(['sudo', 'pacman', '-S', 'nodejs-lts-iron', 'npm', '--noconfirm'], check=True)
    elif package_manager == 'brew':
        subprocess.run(['brew', 'install', 'node'], check=True)
    print("Node.js installation completed")